@@ -10,6 +10,7 @@
 import sys
 
 from pyrseas.dbobject import DbObjectDict, DbSchemaObject
+from pyrseas.dbobject import quote_id, split_schema_table
 from constraint import CheckConstraint, PrimaryKey, ForeignKey, \
     UniqueConstraint
 
@@ -25,7 +26,6 @@ class DbClass(DbSchemaObject):
 class Sequence(DbClass):
     "A sequence generator definition"
 
-    keylist = ['schema', 'name']
     objtype = "SEQUENCE"
 
     def get_attrs(self, dbconn):
@@ -36,12 +36,12 @@ class Sequence(DbClass):
         data = dbconn.fetchone(
             """SELECT start_value, increment_by, max_value, min_value,
                       cache_value
-               FROM %s.%s""" % (self.schema, self.name))
+               FROM %s.%s""" % (quote_id(self.schema), quote_id(self.name)))
         for key, val in data.items():
             setattr(self, key, val)
 
-    def get_owner(self, dbconn):
-        """Get the table and column name that owns the sequence
+    def get_dependent_table(self, dbconn):
+        """Get the table and column name that uses or owns the sequence
 
         :param dbconn: a DbConnection object
         """
@@ -51,12 +51,17 @@ class Sequence(DbClass):
                WHERE objid = '%s'::regclass
                  AND refclassid = 'pg_class'::regclass""" % self.qualname())
         if data:
-            self.owner_table = tbl = data[0]
+            (sch, self.owner_table) = split_schema_table(data[0], self.schema)
             self.owner_column = data[1]
-            if '.' in tbl:
-                dot = tbl.index('.')
-                if self.schema == tbl[:dot]:
-                    self.owner_table = tbl[dot + 1:]
+            return
+        data = dbconn.fetchone(
+            """SELECT adrelid::regclass
+               FROM pg_attrdef a JOIN pg_depend ON (a.oid = objid)
+               WHERE refobjid = '%s'::regclass
+               AND classid = 'pg_attrdef'::regclass""" % self.qualname())
+        if data:
+            (sch, self.dependent_table) = split_schema_table(
+                data[0], self.schema)
 
     def to_map(self):
         """Convert a sequence definition to a YAML-suitable format
@@ -65,7 +70,7 @@ class Sequence(DbClass):
         """
         seq = {}
         for key, val in self.__dict__.items():
-            if key in self.keylist:
+            if key in self.keylist or key == 'dependent_table':
                 continue
             if key == 'max_value' and val == MAX_BIGINT:
                 seq[key] = None
@@ -104,7 +109,8 @@ class Sequence(DbClass):
         if pth:
             stmts.append(pth)
         stmts.append("ALTER SEQUENCE %s OWNED BY %s.%s" % (
-                self.name, self.owner_table, self.owner_column))
+                quote_id(self.name), quote_id(self.owner_table),
+                quote_id(self.owner_column)))
         return stmts
 
     def diff_map(self, inseq):
@@ -169,7 +175,9 @@ class Table(DbClass):
             return
         cols = []
         for i in range(len(self.columns)):
-            cols.append(self.columns[i].to_map())
+            col = self.columns[i].to_map()
+            if col:
+                cols.append(col)
         tbl = {'columns': cols}
         if hasattr(self, 'description'):
             tbl.update(description=self.description)
@@ -204,18 +212,60 @@ class Table(DbClass):
             for k in self.indexes.values():
                 tbl['indexes'].update(self.indexes[k.name].to_map(
                         self.column_names()))
+        if hasattr(self, 'inherits'):
+            if not 'inherits' in tbl:
+                tbl['inherits'] = self.inherits
+        if hasattr(self, 'rules'):
+            if not 'rules' in tbl:
+                tbl['rules'] = {}
+            for k in self.rules.values():
+                tbl['rules'].update(self.rules[k.name].to_map())
+        if hasattr(self, 'triggers'):
+            if not 'triggers' in tbl:
+                tbl['triggers'] = {}
+            for k in self.triggers.values():
+                tbl['triggers'].update(self.triggers[k.name].to_map())
 
         return {self.extern_key(): tbl}
 
     def create(self):
-        """Return a SQL statement to CREATE the table
+        """Return SQL statements to CREATE the table
+
+        :return: SQL statements
+        """
+        stmts = []
+        if hasattr(self, 'created'):
+            return stmts
+        cols = []
+        for col in self.columns:
+            if not (hasattr(col, 'inherited') and col.inherited):
+                cols.append("    " + col.add())
+        inhclause = ''
+        if hasattr(self, 'inherits'):
+            inhclause = " INHERITS (%s)" % ", ".join(t for t in self.inherits)
+        stmts.append("CREATE TABLE %s (\n%s)%s" % (
+                self.qualname(), ",\n".join(cols), inhclause))
+        if hasattr(self, 'description'):
+            stmts.append(self.comment())
+        for col in self.columns:
+            if hasattr(col, 'description'):
+                stmts.append(col.comment())
+        self.created = True
+        return stmts
+
+    def drop(self):
+        """Return a SQL DROP statement for the table
 
         :return: SQL statement
         """
-        cols = []
-        for col in self.columns:
-            cols.append("    " + col.add())
-        return "CREATE TABLE %s (\n%s)" % (self.qualname(), ",\n".join(cols))
+        stmts = []
+        if not hasattr(self, 'dropped') or not self.dropped:
+            if hasattr(self, 'dependent_funcs'):
+                for fnc in self.dependent_funcs:
+                    stmts.append(fnc.drop())
+            self.dropped = True
+            stmts.append("DROP TABLE %s" % self.identifier())
+        return stmts
 
     def diff_map(self, intable):
         """Generate SQL to transform an existing table
@@ -235,6 +285,9 @@ class Table(DbClass):
         base = "ALTER TABLE %s\n    " % self.qualname()
         # check input columns
         for (num, incol) in enumerate(intable.columns):
+            if hasattr(incol, 'oldname'):
+                assert(self.columns[num].name == incol.oldname)
+                stmts.append(self.columns[num].rename(incol.name))
             # add new columns
             if num >= dbcols:
                 stmts.append(base + "ADD COLUMN %s" % incol.add())
@@ -245,6 +298,55 @@ class Table(DbClass):
                 if stmt:
                     stmts.append(base + stmt)
 
+        stmts.append(self.diff_description(intable))
+
+        return stmts
+
+
+class View(DbClass):
+    """A database view definition
+
+    A view is identified by its schema name and view name.
+    """
+
+    objtype = "VIEW"
+
+    def to_map(self):
+        """Convert a view to a YAML-suitable format
+
+        :return: dictionary
+        """
+        return {self.extern_key(): {'definition': self.definition}}
+
+    def create(self, newdefn=None):
+        """Return SQL statements to CREATE the table
+
+        :return: SQL statements
+        """
+        stmts = []
+        defn = newdefn or self.definition
+        if defn[-1:] == ';':
+            defn = defn[:-1]
+        stmts.append("CREATE%s VIEW %s AS\n   %s" % (
+                newdefn and " OR REPLACE" or '', self.qualname(), defn))
+        if hasattr(self, 'description'):
+            stmts.append(self.comment())
+        return stmts
+
+    def diff_map(self, inview):
+        """Generate SQL to transform an existing view
+
+        :param inview: a YAML map defining the new view
+        :return: list of SQL statements
+
+        Compares the view to an input view and generates SQL
+        statements to transform it into the one represented by the
+        input.
+        """
+        stmts = []
+        if self.definition != inview.definition:
+            stmts.append(self.create(inview.definition))
+        stmts.append(self.diff_description(inview))
         return stmts
 
 
@@ -253,14 +355,25 @@ class ClassDict(DbObjectDict):
 
     cls = DbClass
     query = \
-        """SELECT nspname AS schema, relname AS name, relkind AS kind
-           FROM pg_class
+        """SELECT nspname AS schema, relname AS name, relkind AS kind,
+                  CASE WHEN relkind = 'v' THEN pg_get_viewdef(c.oid, TRUE)
+                       ELSE '' END AS definition,
+                  description
+           FROM pg_class c
                 JOIN pg_namespace ON (relnamespace = pg_namespace.oid)
                 JOIN pg_roles ON (nspowner = pg_roles.oid)
-           WHERE relkind in ('r', 'S')
+                LEFT JOIN pg_description d
+                     ON (c.oid = d.objoid AND d.objsubid = 0)
+           WHERE relkind in ('r', 'S', 'v')
                  AND (nspname = 'public' OR rolname <> 'postgres')
            ORDER BY nspname, relname"""
 
+    inhquery = \
+        """SELECT inhrelid::regclass AS sub, inhparent::regclass AS parent,
+                  inhseqno
+           FROM pg_inherits
+           ORDER BY 1, 3"""
+
     def _from_catalog(self):
         """Initialize the dictionary of tables by querying the catalogs"""
         for table in self.fetch():
@@ -272,7 +385,15 @@ class ClassDict(DbObjectDict):
             elif kind == 'S':
                 self[(sch, tbl)] = inst = Sequence(**table.__dict__)
                 inst.get_attrs(self.dbconn)
-                inst.get_owner(self.dbconn)
+                inst.get_dependent_table(self.dbconn)
+            elif kind == 'v':
+                self[(sch, tbl)] = View(**table.__dict__)
+        for (tbl, partbl, num) in self.dbconn.fetchall(self.inhquery):
+            (sch, tbl) = split_schema_table(tbl)
+            table = self[(sch, tbl)]
+            if not hasattr(table, 'inherits'):
+                table.inherits = []
+            table.inherits.append(partbl)
 
     def from_map(self, schema, inobjs, newdb):
         """Initalize the dictionary of tables by converting the input map
@@ -298,11 +419,19 @@ class ClassDict(DbObjectDict):
                 except KeyError, exc:
                     exc.args = ("Table '%s' has no columns" % key, )
                     raise
+                if 'inherits' in intable:
+                    table.inherits = intable['inherits']
                 if 'oldname' in intable:
                     table.oldname = intable['oldname']
                 newdb.constraints.from_map(table, intable)
                 if 'indexes' in intable:
                     newdb.indexes.from_map(table, intable['indexes'])
+                if 'rules' in intable:
+                    newdb.rules.from_map(table, intable['rules'])
+                if 'triggers' in intable:
+                    newdb.triggers.from_map(table, intable['triggers'])
+                if 'description' in intable:
+                    table.description = intable['description']
             elif objtype == 'sequence':
                 self[(schema.name, key)] = seq = Sequence(
                     schema=schema.name, name=key)
@@ -313,36 +442,59 @@ class ClassDict(DbObjectDict):
                     setattr(seq, attr, val)
                 if 'oldname' in inseq:
                     seq.oldname = inseq['oldname']
+            elif objtype == 'view':
+                self[(schema.name, key)] = view = View(
+                    schema=schema.name, name=key)
+                inview = inobjs[k]
+                if not inview:
+                    raise ValueError("View '%s' has no specification" % k)
+                for attr, val in inview.items():
+                    setattr(view, attr, val)
+                if 'oldname' in inview:
+                    view.oldname = inview['oldname']
             else:
                 raise KeyError("Unrecognized object type: %s" % k)
 
-    def link_refs(self, dbcolumns, dbconstrs, dbindexes):
+    def link_refs(self, dbcolumns, dbconstrs, dbindexes, dbrules, dbtriggers):
         """Connect columns, constraints, etc. to their respective tables
 
         :param dbcolumns: dictionary of columns
         :param dbconstrs: dictionary of constraints
         :param dbindexes: dictionary of indexes
+        :param dbrules: dictionary of rules
+        :param dbtriggers: dictionary of triggers
 
         Links each list of table columns in `dbcolumns` to the
         corresponding table. Fills the `foreign_keys`,
-        `unique_constraints` and `indexes` dictionaries for each table
-        by traversing the `dbconstrs` and `dbindexes` dictionaries,
-        which are keyed by schema, table and constraint or index.
+        `unique_constraints`, `indexes` and `triggers` dictionaries
+        for each table by traversing the `dbconstrs`, `dbindexes` and
+        `dbtriggers` dictionaries, which are keyed by schema, table
+        and constraint, index or trigger name.
         """
         for (sch, tbl) in dbcolumns.keys():
             assert self[(sch, tbl)]
             self[(sch, tbl)].columns = dbcolumns[(sch, tbl)]
+            for col in dbcolumns[(sch, tbl)]:
+                col._table = self[(sch, tbl)]
         for (sch, tbl) in self.keys():
-            if isinstance(self[(sch, tbl)], Sequence):
-                seq = self[(sch, tbl)]
-                if hasattr(seq, 'owner_table'):
-                    if isinstance(seq.owner_column, int):
-                        seq.owner_column = self[(sch, seq.owner_table)]. \
-                            column_names()[seq.owner_column - 1]
+            table = self[(sch, tbl)]
+            if isinstance(table, Sequence) and hasattr(table, 'owner_table'):
+                if isinstance(table.owner_column, int):
+                    table.owner_column = self[(sch, table.owner_table)]. \
+                        column_names()[table.owner_column - 1]
+            elif isinstance(table, Table) and hasattr(table, 'inherits'):
+                for partbl in table.inherits:
+                    (parsch, partbl) = split_schema_table(partbl)
+                    assert self[(parsch, partbl)]
+                    parent = self[(parsch, partbl)]
+                    if not hasattr(parent, 'descendants'):
+                        parent.descendants = []
+                    parent.descendants.append(table)
         for (sch, tbl, cns) in dbconstrs.keys():
             constr = dbconstrs[(sch, tbl, cns)]
-            if (sch, tbl) not in self:  # check constraints on domains
+            if hasattr(constr, 'target'):
                 continue
+            assert self[(sch, tbl)]
             table = self[(sch, tbl)]
             if isinstance(constr, CheckConstraint):
                 if not hasattr(table, 'check_constraints'):
@@ -369,6 +521,20 @@ class ClassDict(DbObjectDict):
             if not hasattr(table, 'indexes'):
                 table.indexes = {}
             table.indexes.update({idx: dbindexes[(sch, tbl, idx)]})
+        for (sch, tbl, rul) in dbrules.keys():
+            assert self[(sch, tbl)]
+            table = self[(sch, tbl)]
+            if not hasattr(table, 'rules'):
+                table.rules = {}
+            table.rules.update({rul: dbrules[(sch, tbl, rul)]})
+            dbrules[(sch, tbl, rul)]._table = self[(sch, tbl)]
+        for (sch, tbl, trg) in dbtriggers.keys():
+            assert self[(sch, tbl)]
+            table = self[(sch, tbl)]
+            if not hasattr(table, 'triggers'):
+                table.triggers = {}
+            table.triggers.update({trg: dbtriggers[(sch, tbl, trg)]})
+            dbtriggers[(sch, tbl, trg)]._table = self[(sch, tbl)]
 
     def _rename(self, obj, objtype):
         """Process a RENAME"""
@@ -376,6 +542,7 @@ class ClassDict(DbObjectDict):
         oldname = obj.oldname
         try:
             stmt = self[(obj.schema, oldname)].rename(obj.name)
+            self[(obj.schema, obj.name)] = self[(obj.schema, oldname)]
             del self[(obj.schema, oldname)]
         except KeyError, exc:
             exc.args = ("Previous name '%s' for %s '%s' not found" % (
@@ -408,16 +575,43 @@ class ClassDict(DbObjectDict):
                     stmts.append(inseq.create())
 
         # check input tables
+        inhstack = []
         for (sch, tbl) in intables.keys():
             intable = intables[(sch, tbl)]
             if not isinstance(intable, Table):
                 continue
             # does it exist in the database?
             if (sch, tbl) not in self:
-                if hasattr(intable, 'oldname'):
-                    stmts.append(self._rename(intable, "table"))
-                else:
+                if not hasattr(intable, 'oldname'):
                     # create new table
+                    if hasattr(intable, 'inherits'):
+                        inhstack.append(intable)
+                    else:
+                        stmts.append(intable.create())
+                else:
+                    stmts.append(self._rename(intable, "table"))
+        while len(inhstack):
+            intable = inhstack.pop()
+            createit = True
+            for partbl in intable.inherits:
+                if intables[split_schema_table(partbl)] in inhstack:
+                    createit = False
+            if createit:
+                stmts.append(intable.create())
+            else:
+                inhstack.insert(0, intable)
+
+        # check input views
+        for (sch, tbl) in intables.keys():
+            intable = intables[(sch, tbl)]
+            if not isinstance(intable, View):
+                continue
+            # does it exist in the database?
+            if (sch, tbl) not in self:
+                if hasattr(intable, 'oldname'):
+                    stmts.append(self._rename(intable, "view"))
+                else:
+                    # create new view
                     stmts.append(intable.create())
 
         # second pass: input sequences not owned by tables
@@ -435,14 +629,14 @@ class ClassDict(DbObjectDict):
                     # create new sequence
                     stmts.append(inseq.create())
 
-        # check database tables and sequences
+        # check database tables, sequences and views
         for (sch, tbl) in self.keys():
             table = self[(sch, tbl)]
             # if missing, mark it for dropping
             if (sch, tbl) not in intables:
                 table.dropped = False
             else:
-                # check table/sequence objects
+                # check table/sequence/view objects
                 stmts.append(table.diff_map(intables[(sch, tbl)]))
 
         # now drop the marked tables
@@ -451,7 +645,31 @@ class ClassDict(DbObjectDict):
             if isinstance(table, Sequence) and hasattr(table, 'owner_table'):
                 continue
             if hasattr(table, 'dropped') and not table.dropped:
-                # drop subordinate objects first
+                # first, drop all foreign keys
+                if hasattr(table, 'foreign_keys'):
+                    for fgn in table.foreign_keys:
+                        stmts.append(table.foreign_keys[fgn].drop())
+                # and drop the triggers
+                if hasattr(table, 'triggers'):
+                    for trg in table.triggers:
+                        stmts.append(table.triggers[trg].drop())
+                if hasattr(table, 'rules'):
+                    for rul in table.rules:
+                        stmts.append(table.rules[rul].drop())
+                # drop views
+                if isinstance(table, View):
+                    stmts.append(table.drop())
+
+        inhstack = []
+        for (sch, tbl) in self.keys():
+            table = self[(sch, tbl)]
+            if (isinstance(table, Sequence) \
+                    and (hasattr(table, 'owner_table') \
+                             or hasattr(table, 'dependent_table'))) \
+                             or isinstance(table, View):
+                continue
+            if hasattr(table, 'dropped') and not table.dropped:
+                # next, drop other subordinate objects
                 if hasattr(table, 'check_constraints'):
                     for chk in table.check_constraints:
                         stmts.append(table.check_constraints[chk].drop())
@@ -461,14 +679,34 @@ class ClassDict(DbObjectDict):
                 if hasattr(table, 'indexes'):
                     for idx in table.indexes:
                         stmts.append(table.indexes[idx].drop())
-                if hasattr(table, 'foreign_keys'):
-                    for fgn in table.foreign_keys:
-                        stmts.append(table.foreign_keys[fgn].drop())
+                if hasattr(table, 'rules'):
+                    for rul in table.rules:
+                        stmts.append(table.rules[rul].drop())
                 if hasattr(table, 'primary_key'):
                     # TODO there can be more than one referred_by
                     if hasattr(table, 'referred_by'):
                         stmts.append(table.referred_by.drop())
                     stmts.append(table.primary_key.drop())
+                # finally, drop the table itself
+                if hasattr(table, 'descendants'):
+                    inhstack.append(table)
+                else:
+                    stmts.append(table.drop())
+        while len(inhstack):
+            table = inhstack.pop()
+            dropit = True
+            for childtbl in table.descendants:
+                if self[(childtbl.schema, childtbl.name)] in inhstack:
+                    dropit = False
+            if dropit:
+                stmts.append(table.drop())
+            else:
+                inhstack.insert(0, table)
+        for (sch, tbl) in self.keys():
+            table = self[(sch, tbl)]
+            if isinstance(table, Sequence) \
+                    and hasattr(table, 'dependent_table') \
+                    and hasattr(table, 'dropped') and not table.dropped:
                 stmts.append(table.drop())
 
         # last pass to deal with nextval DEFAULTs
